{-|
Module      : Main
Description : Implementação de um jogo de Damas utilizando ADTs e Zipper 
Copyright   : (c) Fabricio Olivetti de Franca, 2020
License     : GPL-3
Maintainer  : fabricio.olivetti@gmail.com
Stability   : experimental
Portability : POSIX

Código de exemplo do uso de Zippers para percorrer uma estrutura 2D.
O intuito desse código é ilustrar como a estrutura de Zippers combinada
com uma função de visão local facilita a criação de funções auxiliares.
-}

module Main where

import Data.List.Split (chunksOf)
import Data.Array 

-- | Uma peça de dama pode ser Preta (P),
-- Branca (B), ou Vazia (V) ou Proibido (F)  [proibido acrescenta depois]
data Checkers = B | P | V | F deriving Eq

-- | Instância de Show para o tipo Checkers 
instance Show Checkers where
  show B = "⚪"
  show P = "⚫"
  show V = "⬜"
  show F = ""

-- | Estrutura de Zipper de lista
data Board a = Board { foco  :: (Int, Int)          -- coordenada do foco atual 
                     , board :: Array (Int, Int) a  -- dados que contém os focos 
                     } deriving Eq 

-- | Instância de Show para um tabuleiro. Vamos assumir que sempre será 8x8.
instance Show a => Show (Board a) where
  show (Board _ board) = unlines 
                       $ map (unwords . map show) 
                       $ chunksOf 8 
                       $ elems board

-- | Função que atualiza uma coordenada 
type CoordFun  = (Int, Int) -> (Int, Int)

-- | Lista de movimentos que devem ser feitos para atualizar o tabuleiro. 
type Movimento = [((Int,Int), Checkers)]

-- * Conversões

-- | Tabuleiro inicial
initialBoard :: Board Checkers
initialBoard = Board (1,1) $ listArray ((1,1),(8,8)) inicio 
  where
    inicio = concat [vb, bv, vb, v, v, pv, vp, pv]
    padrão = concat . replicate 4

    vb    = padrão [V,B]
    bv    = padrão [B,V]
    pv    = padrão [P,V]
    vp    = padrão [V,P]
    v     = padrão [V,V]

-- * Como caminhar pelo Zipper

-- | Verifica se coordenada está dentro do limite
within :: Ord a => (a,a) -> ((a, a), (a,a)) -> Bool
within (x,y) ((a,b), (c,d)) = inf && sup
  where
    inf = x >= a && y >= b
    sup = x <= c && y <= d

-- | Recupera o valor do Zipper 
valor :: Board a -> a
valor (Board ix b) = b ! ix

{-
-- estou me repetindo, vamos melhorar
paraDireita :: Board a -> Maybe (Board a)
paraDireita (Board (x,y) b)
  | (x+1, y) `within` bounds b = Just (Board (x+1, y) b)
  | otherwise                  = Nothing

paraEsquerda :: Board a -> Maybe (Board a)
paraEsquerda (Board (x,y) b)  
  | (x-1, y) `within` bounds b = Just (Board (x-1, y) b)
  | otherwise                  = Nothing
-}

-- | Atualiza a coordenada segundo uma função e
-- retorna o novo tabuleiro se a coordenada estiver dentro
-- dos limites.
{-
para :: CoordFun -> Board a -> Maybe (Board a)
para f (Board c b)
  | f c `within` bounds b = Just $ Board (f c) b
  | otherwise             = Nothing
-}
para :: (Int -> Int) -> (Int -> Int) -> Board a -> Maybe (Board a)
para f g (Board (x,y) b)
  | c' `within` bounds b = Just $ Board c' b
  | otherwise             = Nothing
  where
    c' = (f x, g y)

-- | Funções para caminhar em todas as direções
paraDireita, paraEsquerda, paraCima, paraBaixo :: Board a -> Maybe (Board a)
paraDireita  = para id (+1) -- (\(x,y) -> (x, y+1))
paraEsquerda = para id (subtract 1) -- (\(x,y) -> (x, y-1))
paraCima     = para (subtract 1) id -- (\(x,y) -> (x-1, y))
paraBaixo    = para (+1) id -- (\(x,y) -> (x+1, y))

-- | volta para o início da linha atual
rebobina :: Board a -> Board a
rebobina (Board (x,_) b) = Board (x,1) b

-- | caminha para a próxima casa, seguindo pela ordem de linhas.
proximo :: Board a -> Maybe (Board a)
proximo b = 
  case paraDireita b of
       Nothing -> case paraBaixo b of
                       Nothing -> Nothing
                       Just b' -> Just $ rebobina b'
       mb      -> mb

-- | recupera uma vizinhança 3x3 em torno da posição atual.        
vizinhanca :: Board Checkers -> [[Checkers]]
vizinhanca (Board (x,y) b) = [[b !? (i,j) | j <- [y-1..y+1]] | i <- [x-1..x+1]]
  where
    b' !? c 
       | c `within` bounds b' = b' ! c
       | otherwise            = F

-- | verifica se é possível fazer um movimento.
-- Originalmente essa função retornava verdadeiro e falso
-- porém percebemos que precisamos devolver a lista de atualizações do tabuleiro.
-- Uma lista vazia representa impossibilidade em fazer tal movimento.
possivelMover :: Checkers -> [[Checkers]] -> Movimento
possivelMover B
 [[B, _, _]
 ,[_, V, _]
 ,[_, _, _]]
   = [((-1,-1),V), ((0,0), B)] 
possivelMover B
 [[_, _, B]
 ,[_, V, _]
 ,[_, _, _]]
   = [((-1,1),V), ((0,0),B)]
possivelMover P 
 [[_, _, _]
 ,[_, V, _]
 ,[_, _, P]]
   = [((1,1),V), ((0,0),P)]
possivelMover P 
 [[_, _, _]
 ,[_, V, _]
 ,[P, _, _]]
   = [((1,-1),V), ((0,0),P)]
possivelMover _ _ = []

-- | similar ao 'possivelMover' porém verifica possíveis ataques. 
possivelAtacar :: Checkers -> [[Checkers]] -> Movimento
possivelAtacar B 
 [[B, _, _]
 ,[_, P, _]
 ,[_, _, V]]
   = [((-1,-1),V),((0,0),V),((1,1),B)]
possivelAtacar B
 [[_, _, B]
 ,[_, P, _]
 ,[V, _, _]]
   = [((-1,1),V),((0,0),V),((1,-1),B)]
possivelAtacar P 
 [[V, _, _]
 ,[_, B, _]
 ,[_, _, P]]
   = [((-1,-1),P),((0,0),V),((1,1),V)]
possivelAtacar P
 [[_, _, V]
 ,[_, B, _]
 ,[P, _, _]]
   = [((-1,1),P),((0,0),V),((1,-1),V)]
possivelAtacar _ _ = []

-- | volta o foco para a primeira casa do tabuleiro 
paraComeco :: Board a -> Board a
paraComeco (Board _ b) = Board (1,1) b 

ataques, movimentos :: Checkers -> Board Checkers -> Board Movimento
ataques peça    = aplica (possivelAtacar peça . vizinhanca)
movimentos peça = aplica (possivelMover peça . vizinhanca)
--ataques peça = fmap (possivelAtacar peça . vizinhanca) . duplica -- não mostrar na aula 

-- | co-co-co-co-co-comonads!!!
-- Dada uma função que retorna um valor de um tipo b em um certo foco do tabuleiro,
-- aplica em todos os possíveis focos. 
aplica :: (Board a -> b) -> Board a -> Board b
aplica f (Board ix b) = 
  let 
    limites = bounds b
    ixs     = indices b
    g i     = (i, f (Board i b))
    b'      = array limites $ map g ixs
  in Board ix b'
      
{-
-- não mostrarei na aula 
duplica :: Board a -> Board (Board a)
duplica (Board _ b) = Board (1,1) listaArray
  where
    boardOfboards = Board (1,1) b
    listaArray    = array limites [(ix, boardOfboards { foco = ix }) | ix <- indices b]
    limites       = bounds b
-}

-- | Dado um tabuleiro de possíveis movimentos, escolhe o primeiro possível.
-- dando preferência para ataques.
escolha :: Board Movimento -> Movimento
escolha b
  | (not . null) movimento   = map (adiciona (foco b)) movimento
  | otherwise = case proximo b of
                     Nothing -> []
                     Just b' -> escolha b'
  where
    movimento          = valor b
    adiciona xy (ij,p) = (xy .+. ij,p)
    (x,y) .+. (i,j)    = (x+i, y+j)

-- | aplica um movimento 
move :: Checkers -> Board Checkers -> Board Checkers
move p b 
  | null atks = b { board = board b // movs }
  | otherwise = b { board = board b // atks }
  where
    atks = escolha $ ataques p b
    movs = escolha $ movimentos p b

-- | cria uma lista infinita de movimentos alternados
alternaVez :: Checkers -> Board Checkers -> [Board Checkers]
alternaVez p b = b' : alternaVez p' b'
  where
    b' = move p b
    p' = if p==B then P else B

-- | Conta a quantidade de certa peça
conta :: Checkers -> Board Checkers -> Int
conta p b = length 
          $ filter (==p) 
          $ elems 
          $ board b

-- | Verifica se o jogo terminou
terminou :: Board Checkers -> Bool
terminou b = conta P b == 0 || conta B b == 0

main :: IO ()
main = do
  print initialBoard
  mapM_ print $ takeWhile (not.terminou) $ alternaVez B initialBoard
